@@ -1,281 +1,300 @@
 """ Object model and loader for the modelDescription.xml """
 
-
-class ModelDescription(object):
-
-    def __init__(self):
-        self.guid = None
-        self.fmiVersion = None
-        self.modelName = None
-        self.description = None
-        self.generationTool = None
-        self.generationDateAndTime = None
-        self.variableNamingConvention = 'flat'
-        self.numberOfContinuousStates = 0
-        self.numberOfEventIndicators = 0
-
-        self.defaultExperiment = None
-
-        self.coSimulation = None
-        self.modelExchange = None
-        self.scheduledExecution = None
-
-        self.buildConfigurations = []
-
-        self.unitDefinitions = []
-        self.typeDefinitions = []
-        self.modelVariables = []
-
-        # model structure
-        self.outputs = []
-        self.derivatives = []
-        self.initialUnknowns = []
-        self.eventIndicators = []
+from typing import List, Union, IO
+from attr import attrs, attrib, Factory
 
 
+@attrs(auto_attribs=True)
 class DefaultExperiment(object):
 
-    def __init__(self):
-        self.startTime = None
-        self.stopTime = None
-        self.tolerance = None
-        self.stepSize = None
+    startTime: str = None
+    stopTime: str = None
+    tolerance: str = None
+    stepSize: str = None
 
 
+@attrs(eq=False)
 class InterfaceType(object):
 
-    def __init__(self):
-        self.modelIdentifier = None
-        self.needsExecutionTool = False
-        self.canBeInstantiatedOnlyOncePerProcess = False
-        self.canGetAndSetFMUstate = False
-        self.canSerializeFMUstate = False
-        self.providesDirectionalDerivative = False
-        self.providesAdjointDerivatives = False
-        self.providesPerElementDependencies = False
+    modelIdentifier = attrib(type=str, default=None)
+    needsExecutionTool = attrib(type=bool, default=False, repr=False)
+    canBeInstantiatedOnlyOncePerProcess = attrib(type=bool, default=False, repr=False)
+    canGetAndSetFMUstate = attrib(type=bool, default=False, repr=False)
+    canSerializeFMUstate = attrib(type=bool, default=False, repr=False)
+    providesDirectionalDerivative = attrib(type=bool, default=False, repr=False)
+    providesAdjointDerivatives = attrib(type=bool, default=False, repr=False)
+    providesPerElementDependencies = attrib(type=bool, default=False, repr=False)
 
-        # FMI 2.0
-        self.canNotUseMemoryManagementFunctions = False
-        self.providesDirectionalDerivative = False
+    # FMI 2.0
+    canNotUseMemoryManagementFunctions = attrib(type=bool, default=False, repr=False)
+    providesDirectionalDerivative = attrib(type=bool, default=False, repr=False)
 
 
+@attrs(eq=False)
 class ModelExchange(InterfaceType):
 
-    def __init__(self):
-        super(ModelExchange, self).__init__()
-        self.completedIntegratorStepNotNeeded = False
+    completedIntegratorStepNotNeeded = attrib(type=bool, default=False, repr=False)
 
 
+@attrs(eq=False)
 class CoSimulation(InterfaceType):
 
-    def __init__(self):
-        super(CoSimulation, self).__init__()
-        self.canHandleVariableCommunicationStepSize = False
-        self.canInterpolateInputs = False
-        self.maxOutputDerivativeOrder = 0
-        self.canRunAsynchronuously = False
+    canHandleVariableCommunicationStepSize = attrib(type=bool, default=False, repr=False)
+    canInterpolateInputs = attrib(type=bool, default=False, repr=False)
+    maxOutputDerivativeOrder = attrib(type=int, default=0, repr=False)
+    canRunAsynchronuously = attrib(type=bool, default=False, repr=False)
+    providesIntermediateUpdate = attrib(type=bool, default=False, repr=False)
+    recommendedIntermediateInputSmoothness = 0
+    canReturnEarlyAfterIntermediateUpdate = attrib(type=bool, default=False, repr=False)
+    fixedInternalStepSize = attrib(type=float, default=None, repr=False)
+    hasEventMode = attrib(type=bool, default=False, repr=False)
 
 
-class ScheduledExecution(CoSimulation):
+@attrs(eq=False)
+class ScheduledExecution(InterfaceType):
 
-    def __init__(self):
-        super(ScheduledExecution, self).__init__()
-
-
-class BuildConfiguration(object):
-
-    def __init__(self):
-        self.modelIdentifier = None
-        self.sourceFileSets = []
-
-    def __repr__(self):
-        return 'BuildConfiguration %s' % self.sourceFileSets
+    maxOutputDerivativeOrder = attrib(type=int, default=0, repr=False)
+    providesIntermediateUpdate = attrib(type=bool, default=False, repr=False)
+    recommendedIntermediateInputSmoothness = attrib(type=int, default=0, repr=False)
 
 
+@attrs(auto_attribs=True, eq=False)
 class PreProcessorDefinition(object):
 
-    def __init__(self):
-        self.name = None
-        self.value = None
-        self.optional = False
-        self.description = None
-
-    def __repr__(self):
-        return '%s=%s' % (self.name, self.value)
+    name: str = None
+    value: str = None
+    optional: bool = False
+    description: str = None
 
 
+@attrs(auto_attribs=True, eq=False)
 class SourceFileSet(object):
 
-    def __init__(self):
-        self.language = None
-        self.preprocessorDefinitions = []
-        self.sourceFiles = []
-        self.includeDirectories = []
-
-    def __repr__(self):
-        return '%s %s' % (self.language, self.sourceFiles)
+    name: str = None
+    language: str = None
+    compiler: str = None
+    compilerOptions: str = None
+    preprocessorDefinitions: List[str] = Factory(list)
+    sourceFiles: List[str] = Factory(list)
+    includeDirectories: List[str] = Factory(list)
 
 
-class ScalarVariable(object):
+@attrs(auto_attribs=True, eq=False)
+class BuildConfiguration(object):
 
-    def __init__(self, name, valueReference):
-        self.name = name
-        self.valueReference = valueReference
-        self.description = None
-
-        self.type = None
-        "One of 'Real', 'Integer', 'Enumeration', 'Boolean', 'String'"
-
-        self.dimensions = []
-        "List of fixed dimensions"
-
-        self.dimensionValueReferences = []
-        "List of value references to the variables that hold the dimensions"
-
-        self.quantitiy = None
-        "Physical quantity"
-
-        self.unit = None
-        "Unit"
-
-        self.displayUnit = None
-        "Default display unit"
-
-        self.relativeQuantity = False
-        "Relative quantity"
-
-        self.min = None
-        "Minimum value"
-
-        self.max = None
-        "Maximum value"
-
-        self.nominal = None
-        "Nominal value"
-
-        self.unbounded = False
-        "Value is unbounded"
-
-        self.start = None
-        "Initial or guess value"
-
-        self.derivative = None
-        "Derivative"
-
-        self.causality = None
-        "One of 'parameter', 'calculatedParameter', 'input', 'output', 'local', 'independent'"
-
-        self.variability = None
-        "One of 'constant', 'fixed', 'tunable', 'discrete' or 'continuous'"
-
-        self.initial = None
-        "One of 'exact', 'approx', 'calculated' or None"
-
-        self.declaredType = None
-
-        self.reinit = False
-        "Can be reinitialized at an event by the FMU"
-
-        self.sourceline = None
-        "Line number in the modelDescription.xml or None if unknown"
-
-    def __repr__(self):
-        return '%s "%s"' % (self.type, self.name)
+    modelIdentifier: str = None
+    sourceFileSets: List[SourceFileSet] = Factory(list)
 
 
+@attrs(eq=False)
 class Dimension(object):
 
-    def __init__(self, **kwargs):
-        self.start = kwargs.get('start')
-        self.valueReference = kwargs.get('valueReference')
-
-
-class SimpleType(object):
-    """ Type Definition """
-
-    def __init__(self, **kwargs):
-        self.name = kwargs.get('name')
-        self.type = kwargs.get('type')
-        self.quantity = kwargs.get('quantity')
-        self.unit = kwargs.get('unit')
-        self.displayUnit = kwargs.get('displayUnit')
-        self.relativeQuantity = kwargs.get('relativeQuantity')
-        self.min = kwargs.get('min')
-        self.max = kwargs.get('max')
-        self.nominal = kwargs.get('nominal')
-        self.unbounded = kwargs.get('unbounded')
-        self.items = kwargs.get('items', [])
-
-    def __repr__(self):
-        return '%s "%s" [%s]' % (self.type, self.name, self.unit)
+    start = attrib(type=str)
+    valueReference = attrib(type=int)
 
 
+@attrs(eq=False)
 class Item(object):
     """ Enumeration Item """
 
-    def __init__(self, **kwargs):
-        self.name = kwargs.get('name')
-        self.value = kwargs.get('value')
-        self.description = kwargs.get('description')
-
-    def __repr__(self):
-        return '%s (%s) %s' % (self.name, self.value, self.description)
+    name = attrib(type=str, default=None)
+    value = attrib(type=str, default=None)
+    description = attrib(type=str, default=None, repr=False)
 
 
+@attrs(eq=False)
+class SimpleType(object):
+    """ Type Definition """
+
+    name = attrib(type=str, default=None)
+    type = attrib(type=str, default=None)
+    quantity = attrib(type=str, default=None, repr=False)
+    unit = attrib(type=str, default=None)
+    displayUnit = attrib(type=str, default=None, repr=False)
+    relativeQuantity = attrib(type=str, default=None, repr=False)
+    min = attrib(type=str, default=None, repr=False)
+    max = attrib(type=str, default=None, repr=False)
+    nominal = attrib(type=str, default=None, repr=False)
+    unbounded = attrib(type=str, default=None, repr=False)
+    items = attrib(type=List[Item], default=Factory(list), repr=False)
+
+
+@attrs(eq=False)
 class Unit(object):
 
-    def __init__(self, name):
-        self.name = name
-        self.baseUnit = None
-        self.displayUnits = []
-
-    def __repr__(self):
-        return '%s' % self.name
+    name = attrib(type=str, default=None)
+    baseUnit = attrib(type=str, default=None, repr=False)
+    displayUnits = attrib(type=List[str], default=Factory(list), repr=False)
 
 
+@attrs(eq=False)
 class BaseUnit(object):
 
-    def __init__(self):
-        self.kg = 0
-        self.m = 0
-        self.s = 0
-        self.A = 0
-        self.K = 0
-        self.mol = 0
-        self.cd = 0
-        self.rad = 0
-        self.factor = 1.0
-        self.offset = 0.0
-
-    def __repr__(self):
-        return "kg=%d, m=%d, s=%d, A=%d, K=%d, mol=%d, cd=%d, rad=%d, factor=%g, offset=%g" % (
-            self.kg, self.m, self.s, self.A, self.K, self.mol, self.cd, self.rad, self.factor, self.offset)
+    kg = attrib(type=int, default=0)
+    m = attrib(type=int, default=0)
+    s = attrib(type=int, default=0)
+    A = attrib(type=int, default=0)
+    K = attrib(type=int, default=0)
+    mol = attrib(type=int, default=0)
+    cd = attrib(type=int, default=0)
+    rad = attrib(type=int, default=0)
+    factor = attrib(type=float, default=1.0)
+    offset = attrib(type=float, default=0.0)
 
 
+@attrs(eq=False)
 class DisplayUnit(object):
 
-    def __init__(self, name):
-        self.name = name
-        self.factor = 1.0
-        self.offset = 0.0
-
-    def __repr__(self):
-        return '%s' % self.name
+    name = attrib(type=str, default=None)
+    factor = attrib(type=float, default=1.0, repr=False)
+    offset = attrib(type=float, default=0.0, repr=False)
 
 
+@attrs(eq=False)
+class ScalarVariable(object):
+
+    name = attrib(type=str)
+
+    valueReference = attrib(type=int, repr=False)
+
+    type = attrib(type=str, default=None)
+    "One of 'Real', 'Integer', 'Enumeration', 'Boolean', 'String'"
+
+    description = attrib(type=str, default=None, repr=False)
+
+    causality = attrib(type=str, default=None, repr=False)
+    "One of 'parameter', 'calculatedParameter', 'input', 'output', 'local', 'independent'"
+
+    variability = attrib(type=str, default=None, repr=False)
+    "One of 'constant', 'fixed', 'tunable', 'discrete' or 'continuous'"
+
+    initial = attrib(type=str, default=None, repr=False)
+    "One of 'exact', 'approx', 'calculated' or None"
+
+    interval = attrib(type=str, default=None, repr=False)
+    "One of 'constant', 'fixed', 'calculated', 'tunable', 'changing', 'countdown', 'triggered' or None"
+
+    canHandleMultipleSetPerTimeInstant = attrib(type=bool, default=True, repr=False)
+
+    intermediateUpdate = attrib(type=bool, default=False, repr=False)
+
+    previous = attrib(type=int, default=None, repr=False)
+
+    # TODO: resolve variables
+    clocks = attrib(type=List[int], default=Factory(list))
+
+    declaredType = attrib(type=SimpleType, default=None, repr=False)
+
+    dimensions = attrib(type=List[Dimension], default=Factory(list))
+    "List of fixed dimensions"
+
+    dimensionValueReferences = attrib(type=List[int], default=Factory(list))
+    "List of value references to the variables that hold the dimensions"
+
+    quantity = attrib(type=str, default=None, repr=False)
+    "Physical quantity"
+
+    unit = attrib(type=str, default=None, repr=False)
+    "Unit"
+
+    displayUnit = attrib(type=str, default=None, repr=False)
+    "Default display unit"
+
+    relativeQuantity = attrib(type=bool, default=False, repr=False)
+    "Relative quantity"
+
+    min = attrib(type=str, default=None, repr=False)
+    "Minimum value"
+
+    max = attrib(type=str, default=None, repr=False)
+    "Maximum value"
+
+    nominal = attrib(type=str, default=None, repr=False)
+    "Nominal value"
+
+    unbounded = attrib(type=bool, default=False, repr=False)
+    "Value is unbounded"
+
+    start = attrib(type=str, default=None, repr=False)
+    "Initial or guess value"
+
+    derivative = attrib(type='ScalarVariable', default=None, repr=False)
+    "The derivative of this variable"
+
+    reinit = attrib(type=bool, default=False, repr=False)
+    "Can be reinitialized at an event by the FMU"
+
+    sourceline = attrib(type=int, default=None, repr=False)
+    "Line number in the modelDescription.xml or None if unknown"
+
+    # Clock attributes
+    # TODO: add type hints
+    canBeDeactivated = None
+    priority = None
+    interval = None
+    intervalDecimal = None
+    shiftDecimal = None
+    supportsFraction = None
+    resolution = None
+    intervalCounter = None
+    shiftCounter = None
+
+
+@attrs(eq=False)
 class Unknown(object):
 
-    def __init__(self):
+    index = attrib(type=int, default=0, repr=False)
+    variable = attrib(type=ScalarVariable, default=None)
+    dependencies = attrib(type=List[ScalarVariable], default=Factory(list), repr=False)
+    dependenciesKind = attrib(type=List[str], default=Factory(list), repr=False)
+    sourceline = attrib(type=int, default=0, repr=False)
+    "Line number in the modelDescription.xml"
 
-        self.index = 0
-        self.variable = None
-        self.dependencies = []
-        self.dependenciesKind = []
-        self.sourceline = None
-        "Line number in the modelDescription.xml"
 
-    def __repr__(self):
-        return '%s' % self.variable
+@attrs(eq=False)
+class ModelDescription(object):
+
+    guid = attrib(type=str, default=None, repr=False)
+    fmiVersion = attrib(type=str, default=None)
+    modelName = attrib(type=str, default=None)
+    description = attrib(type=str, default=None, repr=False)
+    generationTool = attrib(type=str, default=None, repr=False)
+    generationDateAndTime = attrib(type=str, default=None, repr=False)
+    variableNamingConvention = attrib(type=str, default='flat', repr=False)
+    numberOfContinuousStates = attrib(type=int, default=0, repr=False)
+    numberOfEventIndicators = attrib(type=int, default=0, repr=False)
+
+    defaultExperiment = attrib(type=DefaultExperiment, default=None, repr=False)
+
+    coSimulation = attrib(type=CoSimulation, default=None)
+    modelExchange = attrib(type=ModelExchange, default=None)
+    scheduledExecution = attrib(type=ScheduledExecution, default=None)
+
+    buildConfigurations = attrib(type=List[BuildConfiguration], default=Factory(list), repr=False)
+
+    unitDefinitions = attrib(type=List[Unit], default=Factory(list), repr=False)
+    typeDefinitions = attrib(type=List[SimpleType], default=Factory(list), repr=False)
+    modelVariables = attrib(type=List[ScalarVariable], default=Factory(list), repr=False)
+
+    # model structure
+    outputs = attrib(type=List[Unknown], default=Factory(list), repr=False)
+    derivatives = attrib(type=List[Unknown], default=Factory(list), repr=False)
+    clockedStates = attrib(type=List[Unknown], default=Factory(list), repr=False)
+    eventIndicators = attrib(type=List[Unknown], default=Factory(list), repr=False)
+    initialUnknowns = attrib(type=List[Unknown], default=Factory(list), repr=False)
+
+
+class ValidationError(Exception):
+    """ Exception raised for failed validation of the modelDescription.xml
+
+    Attributes:
+        problems    list of problems found
+    """
+
+    def __init__(self, problems):
+        message = "Failed to validate modelDescription.xml:\n\n- " + '\n- '.join(problems)
+        self.problems = problems
+        super(ValidationError, self).__init__(message)
 
 
 def _copy_attributes(element, object, attributes=None):
@@ -309,14 +328,13 @@ def read_build_description(filename, validate=True):
     import zipfile
     from lxml import etree
     import os
-    from .util import _is_string
 
-    if _is_string(filename) and os.path.isdir(filename):  # extracted FMU
+    if isinstance(filename, str) and os.path.isdir(filename):  # extracted FMU
         filename = os.path.join(filename, 'sources/buildDescription.xml')
         if not os.path.isfile(filename):
             return []
         tree = etree.parse(filename)
-    elif _is_string(filename) and os.path.isfile(filename) and filename.lower().endswith('.xml'):  # XML file
+    elif isinstance(filename, str) and os.path.isfile(filename) and filename.lower().endswith('.xml'):  # XML file
         if not os.path.isfile(filename):
             return []
         tree = etree.parse(filename)
@@ -378,7 +396,7 @@ def read_build_description(filename, validate=True):
     return build_configurations
 
 
-def read_model_description(filename, validate=True, validate_variable_names=False, validate_model_structure=False):
+def read_model_description(filename: Union[str, IO], validate: bool = True, validate_variable_names: bool = False, validate_model_structure: bool = False) -> ModelDescription:
     """ Read the model description from an FMU without extracting it
 
     Parameters:
@@ -394,16 +412,16 @@ def read_model_description(filename, validate=True, validate_variable_names=Fals
     import zipfile
     from lxml import etree
     import os
-    from .util import _is_string
     from . import validation
+    import numpy as np
 
     # remember the original filename
     _filename = filename
 
-    if _is_string(filename) and os.path.isdir(filename):  # extracted FMU
+    if isinstance(filename, str) and os.path.isdir(filename):  # extracted FMU
         filename = os.path.join(filename, 'modelDescription.xml')
         tree = etree.parse(filename)
-    elif _is_string(filename) and os.path.isfile(filename) and filename.lower().endswith('.xml'):  # XML file
+    elif isinstance(filename, str) and os.path.isfile(filename) and filename.lower().endswith('.xml'):  # XML file
         tree = etree.parse(filename)
     else:  # FMU as path or file like object
         with zipfile.ZipFile(filename, 'r') as zf:
@@ -433,10 +451,9 @@ def read_model_description(filename, validate=True, validate_variable_names=Fals
             schema = etree.XMLSchema(file=os.path.join(module_dir, 'schema', 'fmi3', 'fmi3ModelDescription.xsd'))
 
         if not schema.validate(root):
-            message = "Failed to validate modelDescription.xml:"
-            for entry in schema.error_log:
-                message += "\n%s (line %d, column %d): %s" % (entry.level_name, entry.line, entry.column, entry.message)
-            raise Exception(message)
+            problems = ["%s (line %d, column %d): %s" % (e.level_name, e.line, e.column, e.message)
+                        for e in schema.error_log]
+            raise ValidationError(problems)
 
     modelDescription = ModelDescription()
     _copy_attributes(root, modelDescription, ['fmiVersion', 'guid', 'modelName', 'description', 'generationTool',
@@ -452,8 +469,6 @@ def read_model_description(filename, validate=True, validate_variable_names=Fals
         modelDescription.numberOfContinuousStates = int(root.get('numberOfContinuousStates'))
     elif is_fmi2:
         modelDescription.numberOfContinuousStates = len(root.findall('ModelStructure/Derivatives/Unknown'))
-    else:
-        modelDescription.numberOfContinuousStates = len(root.findall('ModelStructure/Derivative'))
 
     # default experiment
     for d in root.findall('DefaultExperiment'):
@@ -539,7 +554,7 @@ def read_model_description(filename, validate=True, validate_variable_names=Fals
                 buildConfiguration.sourceFileSets.append(source_file_set)
                 source_file_set.sourceFiles = source_files
 
-    elif is_fmi3 and not (_is_string(_filename) and _filename.endswith('.xml')):
+    elif is_fmi3 and not (isinstance(filename, str) and _filename.endswith('.xml')):
         # read buildDescription.xml if _filename is a folder or ZIP file
         modelDescription.buildConfigurations = read_build_description(_filename, validate=validate)
 
@@ -625,7 +640,7 @@ def read_model_description(filename, validate=True, validate_variable_names=Fals
         'tunable':    {'parameter': 'exact', 'calculatedParameter': 'calculated', 'structuralParameter': 'exact', 'local': 'calculated'},
         'discrete':   {'input': None, 'output': 'calculated', 'local': 'calculated'},
         'continuous': {'input': None, 'output': 'calculated', 'local': 'calculated', 'independent': None},
-        'clock':      {'input': None, 'output': None},
+        'clock':      {'input': 'exact', 'output': 'calculated', 'local': 'calculated'},
     }
 
     # model variables
@@ -639,6 +654,8 @@ def read_model_description(filename, validate=True, validate_variable_names=Fals
         sv.causality = variable.get('causality', default='local')
         sv.variability = variable.get('variability')
         sv.initial = variable.get('initial')
+        sv.interval = variable.get('interval')
+        sv.clocks = variable.get('clocks')
         sv.sourceline = variable.sourceline
 
         if fmiVersion in ['1.0', '2.0']:
@@ -785,7 +802,8 @@ def read_model_description(filename, validate=True, validate_variable_names=Fals
     if is_fmi3:
 
         for attr, element in [(modelDescription.outputs, 'Output'),
-                              (modelDescription.derivatives, 'Derivative'),
+                              (modelDescription.derivatives, 'ContinuousStateDerivative'),
+                              (modelDescription.clockedStates, 'ClockedState'),
                               (modelDescription.initialUnknowns, 'InitialUnknown'),
                               (modelDescription.eventIndicators, 'EventIndicator')]:
 
@@ -807,22 +825,29 @@ def read_model_description(filename, validate=True, validate_variable_names=Fals
 
                 attr.append(unknown)
 
-        modelDescription.numberOfEventIndicators = len(modelDescription.eventIndicators)
-
-        # resolve derivatives
         for variable in modelDescription.modelVariables:
-            if variable.derivative is not None:
-                variable.derivative = modelDescription.modelVariables[int(variable.derivative)]
 
-    problems = []
+            # resolve derivative
+            if variable.derivative is not None:
+                variable.derivative = variables[int(variable.derivative)]
+
+            # resolve clocks
+            if variable.clocks is not None:
+                variable.clocks = [variables[int(vr)] for vr in variable.clocks.strip().split(' ')]
+
+        # calculate numberOfContinuousStates
+        for unknown in modelDescription.derivatives:
+            modelDescription.numberOfContinuousStates += int(np.prod(unknown.variable.shape))
+
+        # calculate numberOfEventIndicators
+        for unknown in modelDescription.eventIndicators:
+            modelDescription.numberOfEventIndicators += int(np.prod(unknown.variable.shape))
 
     if validate:
-        problems += validation.validate_model_description(modelDescription,
-                                                          validate_variable_names=validate_variable_names,
-                                                          validate_model_structure=validate_model_structure)
-
-    if problems:
-        message = ("Failed to validate model description. %d problems were found:\n\n- " % len(problems)) + '\n- '.join(problems)
-        raise Exception(message)
+        problems = validation.validate_model_description(modelDescription,
+                                                         validate_variable_names=validate_variable_names,
+                                                         validate_model_structure=validate_model_structure)
+        if problems:
+            raise ValidationError(problems)
 
     return modelDescription
